import asyncio
import logging
from dataclasses import dataclass
from queue import Empty as QueueEmpty
from typing import Callable, Generic, Iterator, List, Optional, TypeVar

import ray

from levanter.utils.ray_utils import ExceptionInfo, ser_exc_info


T = TypeVar("T")

logger = logging.getLogger(__name__)


@dataclass
class _PrefetchException:
    info: ExceptionInfo


class _Sentinel:
    pass


_SENTINEL = _Sentinel()


class RayPrefetchQueue(Generic[T]):
    def __init__(
        self, producer: Callable[[], Iterator[T]], max_queue_size: int = 100, producer_options: dict | None = None
    ):
        self.max_queue_size = max_queue_size
        if producer_options is None:
            producer_options = {}
        self.queue_actor = _QueueActor.remote(max_queue_size)  # type: ignore
        self.producer_task = _run_producer.options(**producer_options).remote(self.queue_actor, producer)
        self._stopped = False
        self._finished = False

    def queue_size(self):
        return ray.get(self.queue_actor.qsize.remote())

    def __next__(self):
        return self.get_next()

    def __iter__(self):
        return self

    def get_next(self, timeout: float | None = None) -> T:
        """
        Get the next item from the producer. If the producer raises an exception, it will be reraised here.

        If the producer is done, this will raise StopIteration.

        Args:
            timeout (float|None): Timeout in seconds for getting the next item. If None, will block indefinitely.

        Raises:
            Empty: If the queue is empty and the timeout is reached.
        """
        if self._finished:
            raise StopIteration
        # time_in = time.time()
        item = ray.get(self.queue_actor.get_next.remote(timeout))
        # time_out = time.time()
        # if time_out - time_in > 0.1:
        #     current_name = ray.get_runtime_context().get_actor_name()
        #     print(f"{current_name} :: Queue get took {time_out - time_in} seconds :: {self.queue_size()}")
        #     logger.info(f"{current_name} :: Queue get took {time_out - time_in} seconds :: {self.queue_size()}")
        if isinstance(item, _PrefetchException):
            item.info.reraise()
        if isinstance(item, _Sentinel):
            self._finished = True
            raise StopIteration
        return item

    def stop(self):
        ray.cancel(self.producer_task)
        ray.get(self.queue_actor.stop.remote())
        self._stopped = True

    def is_stopped(self):
        return self._stopped

    def drain_available(self, max_size: int) -> List[T]:
        return ray.get(self.queue_actor.drain_available.remote(max_size))


@ray.remote
class _QueueActor:
    def __init__(self, max_queue_size: int):
        self.queue: asyncio.Queue = asyncio.Queue(maxsize=max_queue_size)
        self._stopped = False
        self._finished = False

    async def put(self, item):
        await self.queue.put(item)

    async def get_next(self, timeout: Optional[float] = None):
        try:
            if timeout is not None:
                item = await asyncio.wait_for(self.queue.get(), timeout)
            else:
                item = await self.queue.get()
            if isinstance(item, _Sentinel):
                self._finished = True
            return item
        except asyncio.TimeoutError:
            raise QueueEmpty

    async def drain_available(self, max_size: int) -> List[T]:
        items: list[T] = []
        while len(items) < max_size:
            try:
                item = self.queue.get_nowait()
                if isinstance(item, _Sentinel):
                    self._finished = True
                    break
                if isinstance(item, _PrefetchException):
                    item.info.reraise()
                items.append(item)
            except asyncio.QueueEmpty:
                break
        return items

    async def qsize(self):
        return self.queue.qsize()

    async def stop(self):
        self._stopped = True


@ray.remote
def _run_producer(queue_actor, producer_fn: Callable[[], Iterator[T]]):
    async def _run_producer(queue_actor, producer_fn):
        previous_put = None
        try:
            producer = producer_fn()
            del producer_fn

            while True:
                next_item = next(producer)
                if previous_put is not None:
                    await previous_put
                previous_put = queue_actor.put.remote(next_item)
        except StopIteration:
            if previous_put is not None:
                await previous_put
            await queue_actor.put.remote(_SENTINEL)
        except Exception as e:
            if previous_put is not None:
                await previous_put
            await queue_actor.put.remote(_PrefetchException(ser_exc_info(e)))

    asyncio.run(_run_producer(queue_actor, producer_fn))



from datetime import timedelta

import pytimeparse


def parse_timedelta(td_str) -> timedelta:
    td = timedelta(seconds=pytimeparse.parse(td_str))
    if td.total_seconds() < 0:
        raise ValueError("Cannot encode negative timedelta")  # not worth the trouble

    return td


def encode_timedelta(td: timedelta) -> str:
    """Encodes a timedelta as a string that can be parsed by parse_timedelta/pytimeparse."""
    out = ""
    if td.total_seconds() < 0:
        raise ValueError("Cannot encode negative timedelta")  # not worth the trouble

    if td.days:
        out += f"{td.days}d"

    seconds: float = td.seconds

    if seconds > 3600:
        hours = seconds // 3600
        seconds -= hours * 3600
        out += f"{hours}h"
    if seconds > 60:
        minutes = seconds // 60
        seconds -= minutes * 60
        out += f"{minutes}m"

    if td.microseconds:
        seconds += td.microseconds / 1e6

    if seconds:
        out += f"{seconds}s"

    assert parse_timedelta(out) == td, f"Failed to encode {td} as {out}"
    return out



import abc
from dataclasses import dataclass
from typing import Generic, Optional, Type, TypeVar

import draccus
import equinox as eqx
import jax.numpy as jnp
from jax.random import PRNGKey

import haliax as hax
from haliax import Axis, NamedArray

from levanter.models.attention import AttentionMask
from levanter.models.loss import next_token_loss


LmConfigT = TypeVar("LmConfigT", bound="LmConfig")
LmT = TypeVar("LmT", bound="LmHeadModel")


class LmExample(eqx.Module):
    tokens: hax.NamedArray
    loss_mask: hax.NamedArray
    attn_mask: AttentionMask | NamedArray = AttentionMask.causal()

    @staticmethod
    def causal(
        tokens: hax.NamedArray, *, loss_mask: Optional[hax.NamedArray] = None, ignore_id: Optional[int] = None
    ) -> "LmExample":
        if tokens.ndim != 1:
            raise ValueError("tokens must be a 1D array")

        if not jnp.issubdtype(tokens.dtype, jnp.integer):
            raise ValueError("tokens must be an integer array")

        Pos = tokens.axes[0]

        # don't predict the last token.
        if loss_mask is None:
            loss_mask = 1 - hax.nn.one_hot(-1, Pos, dtype=jnp.float32)

        if ignore_id is not None:
            # we don't compute loss for any tokens matching the ignore index
            ignore_mask = hax.roll(tokens, -1, Pos) != ignore_id
            loss_mask = loss_mask * ignore_mask

        attn_mask = AttentionMask.causal()
        return LmExample(tokens=tokens, loss_mask=loss_mask, attn_mask=attn_mask)


# TODO: for some reason, mypy doesn't like the discover_packages_path argument?
@dataclass(frozen=True)
class LmConfig(draccus.PluginRegistry, abc.ABC, Generic[LmT], discover_packages_path="levanter.models"):  # type: ignore
    @property
    @abc.abstractmethod
    def model_type(cls) -> Type[LmT]:
        pass

    @property
    @abc.abstractmethod
    def KeyPos(self) -> Axis:
        pass

    @property
    @abc.abstractmethod
    def Pos(self) -> Axis:
        pass

    @property
    @abc.abstractmethod
    def Embed(self) -> Axis:
        pass

    cross_entropy_block_size: Optional[int] = None
    """
    The block size for computing cross-entropy loss. This is the number of tokens that are processed together
    in a single block. This can be adjusted to fit within memory constraints. It's deliberately set to a large
    value because it usually faster to compute the loss in larger blocks.
    """

    def flops_per_token(self, vocab_size: int) -> Optional[float]:
        return None

    def build(self, Vocab: Axis, *, key: PRNGKey) -> "LmT":
        return self.model_type.init(Vocab, self, key=key)  # type: ignore


class LmHeadModel(eqx.Module, Generic[LmConfigT]):
    """
    Superclass for models with a language modeling head.
    """

    @property
    @abc.abstractmethod
    def config(self) -> LmConfigT:
        pass

    @property
    @abc.abstractmethod
    def Vocab(self) -> Axis:
        pass

    @property
    def Pos(self) -> Axis:
        return self.config.Pos

    @property
    def KeyPos(self) -> Axis:
        return self.config.KeyPos

    @property
    def Embed(self) -> Axis:
        return self.config.Embed

    @classmethod
    @abc.abstractmethod
    def init(cls, Vocab: Axis, config: LmConfigT, *, key: PRNGKey) -> "LmHeadModel[LmConfigT]":
        pass

    def __call__(
        self, input_ids: NamedArray, attn_mask: Optional[AttentionMask | NamedArray] = None, *, key=None
    ) -> NamedArray:
        """
        Compute the logits for the next token in a sequence.
        Args:
            input_ids: token IDs with shape [..., Pos]
            attn_mask: attention mask with shape [..., Pos, KeyPos]
            key: PRNGKey for random number generation

        Returns:
            NamedArray: logits with shape [..., Pos, Vocab]

        """
        x = self.activations(input_ids, attn_mask, key=key)
        lm_logits = hax.dot(x, self.get_lm_head(), axis=self.Embed)

        return lm_logits

    @abc.abstractmethod
    def activations(
        self, input_ids: NamedArray, attn_mask: Optional[AttentionMask | NamedArray] = None, *, key=None
    ) -> NamedArray:
        """
        Compute the activations for the next token in a sequence.
        Args:
            input_ids: token IDs with shape {Pos}
            attn_mask: attention mask with shape {Pos, KeyPos}
            key: PRNGKey for random number generation

        Returns:
            NamedArray: activations with shape {Pos, Embed}

        """
        pass

    @abc.abstractmethod
    def get_lm_head(self) -> hax.NamedArray:
        """
        The language modeling head of the model. Should have shape {Embed, Vocab}.
        """
        raise NotImplementedError("get_lm_head not implemented")

    @abc.abstractmethod
    def resize_vocab(self, new_size: int, key: Optional[PRNGKey] = None) -> "LmHeadModel[LmConfigT]":
        """
        Resizes the vocabulary of the model. Key may be provided to use random initialization, otherwise, there
        should be some deterministic initialization of any new parameters.
        """
        pass

    @property
    def vocab_size(self) -> int:
        return self.Vocab.size


def compute_next_token_loss(
    model: LmHeadModel,
    example: LmExample,
    *,
    key=None,
    reduction: Optional[hax.ReductionFunction] = hax.mean,
    reduction_axis: Optional[hax.AxisSelection] = None,
    logsumexp_weight: Optional[float] = None,
    loss_dtype: Optional[Type[jnp.dtype]] = jnp.float32,
) -> jnp.ndarray | NamedArray:
    """
    Computes the cross-entropy loss for a language modeling example. If reduction is not None, the loss is reduced
    across the reduction axis (with reduction_axis=None meaning all axes). If reduction is None, the loss is not
    reduced, and the result is a named array with axes (*batch axes, sequence_length).
    """
    activations = model.activations(example.tokens, example.attn_mask, key=key)

    loss = next_token_loss(
        model.Pos,
        model.Embed,
        model.Vocab,
        activations,
        model.get_lm_head(),
        example.tokens,
        loss_mask=example.loss_mask,
        reduction=reduction,
        reduction_axis=reduction_axis,
        logsumexp_weight=logsumexp_weight,
        dtype=loss_dtype,
        block_size=model.config.cross_entropy_block_size,
    )

    return loss



import functools
import os

import equinox as eqx
import equinox.nn as nn
import jax
import jax.numpy as jnp
import numpy as np
from chex import assert_trees_all_close

import levanter
import levanter.optim.sophia


def test_sophia_h():
    key = jax.random.PRNGKey(0)
    model = nn.Linear(4, 4, use_bias=False, key=key)
    data = np.load(f"{os.path.dirname(__file__)}/data/hero_data.npy").astype("float32")
    optimizer = levanter.optim.sophia.sophia_h(
        lr=1,
        b1=0,
        b2=0.99,
        gamma=2,
        weight_decay=0.0,
        clip_threshold=1,
        key=key,
        update_interval=1,
    )
    model = jax.tree_util.tree_map(lambda x: jnp.ones_like(x), model)
    zero_grad = jax.tree_util.tree_map(lambda x: jnp.zeros_like(x), model)

    opt_state = optimizer.init(model)

    def loss_fn(model, data):
        out = eqx.filter_vmap(model)(data)
        return jnp.mean(out**2) * 4

    jit_update = eqx.filter_jit(optimizer.update)

    obj_fn = functools.partial(loss_fn, data=data)
    for i in range(1000):
        _, opt_state = jit_update(zero_grad, opt_state, params=model, obj_fn=obj_fn)

    # print('Test-estimated hessian: most coordinates should be approximately 2')
    # print('Estimated hessian:', opt_state[0].h.weight)
    assert_trees_all_close(opt_state[0].h.weight, 2, rtol=0.2, atol=0.3)  # this is very approximate

    grad_loss_fn = eqx.filter_jit(eqx.filter_value_and_grad(loss_fn))

    loss, grad = grad_loss_fn(model, data)
    model_updates, opt_state = optimizer.update(grad, opt_state, params=model, obj_fn=obj_fn)
    model = eqx.apply_updates(model, model_updates)

    assert_trees_all_close(loss, 15.74834156036377, rtol=1e-3, atol=1e-3)

    # print("Test-model param after 1 step: most coordinates should be very loosely 0.5")
    assert_trees_all_close(model.weight, 0.5, rtol=0.2, atol=0.1)  # this is very approximate

    # print("Test-loss: loss should shrink by approximately 75% after each iteration")
    for i in range(10):
        loss, grad = grad_loss_fn(model, data)
        model_updates, opt_state = optimizer.update(grad, opt_state, params=model, obj_fn=obj_fn)
        model = eqx.apply_updates(model, model_updates)

        # print('Step:', i , "Loss:", loss.item())
        assert loss < 15.74834156036377 * 0.75 ** (i + 1)



import asyncio
from typing import Optional, Sequence

import jax
import numpy as np
from jax.sharding import Mesh

import haliax
from haliax import Axis
from haliax.partitioning import ResourceAxis

from levanter.data.dataset import AsyncDataset, ListAsyncDataset
from levanter.data.loader import DataLoader, check_sharded_consistency

from .test_utils import skip_if_not_enough_devices


def _small_dataset(seq_len=128, num_sequences=200) -> AsyncDataset[Sequence[int]]:
    sequences = [np.arange(seq_len) + 1000 * i for i in range(num_sequences)]

    return ListAsyncDataset(sequences, is_complete=True)


@skip_if_not_enough_devices(2)
def test_local_batched_data_loading_model_axis_2():
    devices = jax.devices()
    model_axis_size = 2

    mesh = Mesh(
        np.array(devices).reshape(-1, model_axis_size),
        (ResourceAxis.DATA, ResourceAxis.MODEL),
    )
    with mesh, haliax.axis_mapping({"batch": ResourceAxis.DATA}):

        seq_len = 128
        cache = _small_dataset(seq_len)
        Batch = Axis("batch", len(devices))
        loader = DataLoader(Batch, cache, max_buffered_batches=10, mesh=mesh, axis_resources=None)

        batches = list(loader)
        for batch in batches:
            check_sharded_consistency(batch, check_disjoint_indices_are_different=True)


def test_local_batched_data_loading_model_axis_1():
    devices = jax.devices()
    model_axis_size = 1

    mesh = Mesh(
        np.array(devices).reshape(-1, model_axis_size),
        (ResourceAxis.DATA, ResourceAxis.MODEL),
    )
    with mesh, haliax.axis_mapping({"batch": ResourceAxis.DATA}):

        seq_len = 128
        cache = _small_dataset(seq_len)
        Batch = Axis("batch", len(devices))
        loader = DataLoader(Batch, cache, max_buffered_batches=10, mesh=mesh, axis_resources=None)

        batches = list(loader)
        for batch in batches:
            check_sharded_consistency(batch, check_disjoint_indices_are_different=True)


class StructuredDataset(AsyncDataset):
    def __init__(self, seq_len):
        super().__init__()
        self.seq_len = seq_len
        self.begin = 0
        self.end = 256
        self.stride = 1

    async def async_len(self) -> int:
        return (self.end - self.begin) // self.stride

    async def getitem_async(self, index: int) -> dict:
        index = self.begin + index * self.stride
        return {
            "input_ids": np.arange(self.seq_len, dtype=np.int32) + index * 1000,
            "labels": np.arange(self.seq_len, dtype=np.int32) + index * 1000,
            "extra": {
                "input_ids": np.arange(self.seq_len, dtype=np.int32) + index * 1000,
                "mask": np.arange(self.seq_len * 2, dtype=np.int32).reshape(-1, 2) + index * 1000,
            },
        }

    async def final_length_is_known(self) -> bool:
        return True

    def is_finite(self) -> bool:
        return True

    async def current_len(self) -> Optional[int]:
        return await self.async_len()

    async def get_batch(self, indices: Sequence[int]):
        out = await asyncio.gather(*(self.getitem_async(i) for i in indices))
        return out


def test_structured_batches_model_axis_1():
    devices = jax.devices()
    model_axis_size = 1

    mesh = Mesh(
        np.array(devices).reshape(-1, model_axis_size),
        (ResourceAxis.DATA, ResourceAxis.MODEL),
    )
    with mesh, haliax.axis_mapping({"batch": ResourceAxis.DATA}):
        seq_len = 128
        dataset = StructuredDataset(seq_len)
        Batch = Axis("batch", len(devices))
        loader = DataLoader(Batch, dataset, max_buffered_batches=10, mesh=mesh, axis_resources=None)

        batches = list(loader)
        for batch in batches:
            check_sharded_consistency(batch, check_disjoint_indices_are_different=True)


@skip_if_not_enough_devices(2)
def test_structured_batches_model_axis_2():
    devices = jax.devices()
    model_axis_size = 2

    mesh = Mesh(
        np.array(devices).reshape(-1, model_axis_size),
        (ResourceAxis.DATA, ResourceAxis.MODEL),
    )
    with mesh, haliax.axis_mapping({"batch": ResourceAxis.DATA}):
        seq_len = 128
        dataset = StructuredDataset(seq_len)
        Batch = Axis("batch", len(devices))
        loader = DataLoader(Batch, dataset, max_buffered_batches=10, mesh=mesh, axis_resources=None)

        batches = list(loader)
        for batch in batches:
            check_sharded_consistency(batch, check_disjoint_indices_are_different=True)


class StructuredDatasetWithNames(AsyncDataset):
    def __init__(self, Height: Axis, Width: Axis, begin, end, stride):
        super().__init__()
        self.Height = Height
        self.Width = Width
        self.begin = begin
        self.end = end
        self.stride = stride

    async def final_length_is_known(self) -> bool:
        return True

    def is_finite(self) -> bool:
        return True

    async def current_len(self) -> Optional[int]:
        return True

    async def get_batch(self, indices: Sequence[int]):
        out = await asyncio.gather(*(self.getitem_async(i) for i in indices))
        return out

    async def async_len(self) -> int:
        return (self.end - self.begin) // self.stride

    async def getitem_async(self, index: int) -> dict:
        index = self.begin + index * self.stride
        return {
            "input_ids": self._gen_image(index),
            "labels": self._gen_image(index),
            "extra": {
                "input_ids": self._gen_image(index),
                "mask": haliax.arange(self.Height) + index * 1000,
            },
        }

    def _gen_image(self, index):
        image = (
            np.arange(self.Height.size * self.Width.size, dtype=np.int32).reshape(self.Height.size, self.Width.size)
            + index * 1000
        )

        return haliax.named(image, (self.Height, self.Width))

    def __iter__(self):
        for i in range(self.begin, self.end, self.stride):
            yield self[i]


def test_structured_batches_model_axis_1_with_names():
    devices = jax.devices()
    model_axis_size = 1

    mesh = Mesh(
        np.array(devices).reshape(-1, model_axis_size),
        (ResourceAxis.DATA, ResourceAxis.MODEL),
    )
    with mesh, haliax.axis_mapping({"batch": ResourceAxis.DATA}):
        Height = Axis("Height", 16)
        Width = Axis("Width", 16)
        dataset = StructuredDatasetWithNames(Height, Width, 0, len(devices) * 10, 1)
        Batch = Axis("batch", len(devices))
        loader = DataLoader(Batch, dataset, max_buffered_batches=10, mesh=mesh, axis_resources=None)

        batches = list(loader)
        for batch in batches:
            check_sharded_consistency(batch, check_disjoint_indices_are_different=True)

        assert len(batches) == 10


@skip_if_not_enough_devices(2)
def test_structured_batches_model_axis_2_with_names():
    devices = jax.devices()
    model_axis_size = 2

    mesh = Mesh(
        np.array(devices).reshape(-1, model_axis_size),
        (ResourceAxis.DATA, ResourceAxis.MODEL),
    )
    with mesh, haliax.axis_mapping({"batch": ResourceAxis.DATA}):
        Height = Axis("Height", 16)
        Width = Axis("Width", 16)
        dataset = StructuredDatasetWithNames(Height, Width, 0, 256, 1)
        Batch = Axis("batch", len(devices))
        loader = DataLoader(Batch, dataset, max_buffered_batches=10, mesh=mesh, axis_resources=None)

        batches = list(loader)
        for batch in batches:
            check_sharded_consistency(batch, check_disjoint_indices_are_different=True)


@skip_if_not_enough_devices(4)
def test_structured_batches_model_axis_2_subsharded():
    """This tests data loading if individual datums are sharded too"""
    devices = jax.devices()
    model_axis_size = 2

    mesh = Mesh(
        np.array(devices).reshape(-1, model_axis_size),
        (ResourceAxis.DATA, ResourceAxis.MODEL),
    )
    Height = Axis("Height", 16)
    Width = Axis("Width", 16)
    with mesh, haliax.axis_mapping({"batch": ResourceAxis.DATA, Height.name: ResourceAxis.MODEL}):
        dataset = StructuredDatasetWithNames(Height, Width, 0, 256, 1)
        Batch = Axis("batch", len(devices))
        loader = DataLoader(Batch, dataset, max_buffered_batches=10, mesh=mesh, axis_resources=None)

        for batch in iter(loader):
            check_sharded_consistency(batch, check_disjoint_indices_are_different=True)



import dataclasses
from dataclasses import dataclass
from functools import partial
from typing import Callable, Dict, Optional, Type

import equinox as eqx
import jax.numpy as jnp
import jax.random as jrandom
from jaxtyping import PRNGKeyArray

import haliax as hax
import haliax.jax_utils
import haliax.nn as hnn
from haliax import Axis, NamedArray
from haliax.jax_utils import named_call, shaped_rng_split
from haliax.nn.scan import Stacked
from haliax.state_dict import ModuleWithStateDictSerialization

from levanter.compat.hf_checkpoints import HFCheckpointConverter, HFCompatConfig, LmWithHfSerializationMixin
from levanter.logging import silence_transformer_nag
from levanter.models.attention import AttentionBackend, AttentionMask, dot_product_attention
from levanter.models.lm_model import LmConfig
from levanter.utils.flop_utils import lm_flops_per_token


silence_transformer_nag()
from transformers import GPT2Config as HfGpt2Config  # noqa: E402
from transformers import PretrainedConfig as HfConfig  # noqa: E402


@LmConfig.register_subclass("gpt2")
@dataclass(frozen=True)
class Gpt2Config(HFCompatConfig):
    seq_len: int = 1024
    hidden_dim: int = 768
    num_layers: int = 12
    num_heads: int = 12

    # how much to scale the embedding dim for the mlp layer
    mlp_scale: int = 4

    initializer_range: float = 0.02
    # dropout doesn't really help so we 0 it out by default
    embed_pdrop: float = 0.0
    resid_pdrop: float = 0.0
    attn_pdrop: float = 0.0
    layer_norm_epsilon: float = 1e-5
    activation_function: str = "gelu_new"

    # mistral tweaks:
    scale_attn_by_inverse_layer_idx: bool = False
    upcast_attn: bool = False

    gradient_checkpointing: bool = True  # better to just always use this
    gradient_checkpointing_block_size: int = 5

    use_bias: bool = True

    use_flash_attention: Optional[bool] = None
    attn_backend: Optional[AttentionBackend] = None
    flash_attention_block_size: Optional[int] = None

    # Axes
    Pos = property(lambda self: Axis(name="position", size=self.seq_len))
    KeyPos = property(lambda self: self.Pos.alias("key_position"))
    Embed = property(lambda self: Axis(name="embed", size=self.hidden_dim))
    Heads = property(lambda self: Axis(name="heads", size=self.num_heads))
    Layers = property(lambda self: Axis(name="layers", size=self.num_layers))
    Mlp = property(lambda self: Axis(name="mlp", size=self.hidden_dim * self.mlp_scale))
    HeadSize = property(lambda self: Axis(name="head_size", size=self.hidden_dim // self.num_heads))

    @property
    def model_type(self) -> Type["Gpt2LMHeadModel"]:
        return Gpt2LMHeadModel

    def hf_checkpoint_converter(self) -> HFCheckpointConverter["Gpt2Config"]:  # type: ignore
        # We trust this code because it's in our hub repo
        return HFCheckpointConverter(self.__class__, reference_checkpoint="gpt2", ignore_prefix="transformer")

    def to_hf_config(self, vocab_size, config_overrides=None) -> HfGpt2Config:
        if config_overrides is None:
            config_overrides = {}

        return HfGpt2Config(
            vocab_size=vocab_size,
            n_positions=self.seq_len,
            n_layer=self.num_layers,
            n_head=self.num_heads,
            n_embd=self.hidden_dim,
            initializer_range=self.initializer_range,
            attn_pdrop=self.attn_pdrop,
            embd_pdrop=self.embed_pdrop,
            layer_norm_epsilon=self.layer_norm_epsilon,
            activation_function=self.activation_function,
            scale_attn_by_inverse_layer_idx=self.scale_attn_by_inverse_layer_idx,
            reorder_and_upcast_attn=self.upcast_attn,
            **config_overrides,
        )

    @classmethod
    def from_hf_config(cls, hf_config: HfConfig):
        return Gpt2Config(
            seq_len=hf_config.n_positions,
            # vocab_size=config.vocab_size,
            num_layers=hf_config.n_layer,
            num_heads=hf_config.n_head,
            hidden_dim=hf_config.n_embd,
            initializer_range=hf_config.initializer_range,
            attn_pdrop=hf_config.attn_pdrop,
            embed_pdrop=hf_config.embd_pdrop,
            layer_norm_epsilon=hf_config.layer_norm_epsilon,
            activation_function=hf_config.activation_function,
            scale_attn_by_inverse_layer_idx=hf_config.scale_attn_by_inverse_layer_idx,
            upcast_attn=hf_config.reorder_and_upcast_attn,
        )

    def flops_per_token(self, vocab_size: int) -> Optional[float]:
        return lm_flops_per_token(
            hidden_dim=self.hidden_dim,
            intermediate_dim=self.hidden_dim * self.mlp_scale,
            num_layers=self.num_layers,
            num_kv_heads=self.num_heads,
            num_heads=self.num_heads,
            seq_len=self.seq_len,
            vocab_size=vocab_size,
            glu=False,
        )


class Gpt2Mlp(eqx.Module):
    c_fc: hnn.Linear  # projection from Embed to Intermediate (typically 4x Embed)
    c_proj: hnn.Linear  # projection from Intermediate to Embed
    act: Callable = eqx.static_field()

    @staticmethod
    def init(Embed: Axis, Mlp: Axis, activation_fn, *, key, use_bias: bool = True) -> "Gpt2Mlp":
        k_fc, k_proj = jrandom.split(key, 2)
        c_fc = hnn.Linear.init(Out=Mlp, In=Embed, key=k_fc, use_bias=use_bias, out_first=False)
        c_proj = hnn.Linear.init(Out=Embed, In=Mlp, key=k_proj, use_bias=use_bias, out_first=False)
        if isinstance(activation_fn, str):
            activation_fn = ACT2FN[activation_fn]
        act = activation_fn  # type: ignore

        return Gpt2Mlp(c_fc, c_proj, act)

    @named_call
    def __call__(self, x: NamedArray, *, key=None):
        k1, k2 = haliax.jax_utils.maybe_rng_split(key, 2)
        x = self.c_fc(x, key=k1)
        x = self.act(x)
        x = self.c_proj(x, key=k2)
        return x


class Gpt2Attention(eqx.Module):
    config: Gpt2Config = eqx.static_field()

    c_attn: hnn.Linear  # input projection from [embed] -> [(q, k, v), heads, head_dim]
    c_proj: hnn.Linear  # output projection from [heads, head_dim] -> [embed]
    inference: bool

    @staticmethod
    def init(config: Gpt2Config, *, key) -> "Gpt2Attention":
        Qkv = Axis("qkv", size=3)
        use_bias = config.use_bias
        Embed = config.Embed

        k_c, k_proj = jrandom.split(key, 2)
        c_attn = hnn.Linear.init(
            In=Embed, Out=(Qkv, config.Heads, config.HeadSize), key=k_c, use_bias=use_bias, out_first=False
        )
        c_proj = hnn.Linear.init(
            In=(config.Heads, config.HeadSize), Out=Embed, key=k_proj, use_bias=use_bias, out_first=False
        )

        return Gpt2Attention(config, c_attn, c_proj, inference=False)

    @named_call
    def __call__(self, x: NamedArray, mask: Optional[AttentionMask | NamedArray], layer_idx, *, key):
        k_drop, k_attn, k_out = hax.jax_utils.maybe_rng_split(key, 3)
        qkv_out = self.c_attn(x, key=k_attn).rearrange((..., "qkv", "heads", "position", "head_size"))
        q, k, v = qkv_out.unbind("qkv")

        # Rename k and v's Pos as haliax doesn't support unnamed axes or duplicate axes
        k = k.rename({"position": "key_position"})
        v = v.rename({"position": "key_position"})

        # mistral tweak: attention scores can overflow FP16, or just be too imprecise, so upcast to FP32
        if self.config.scale_attn_by_inverse_layer_idx:
            q = q / (layer_idx + 1.0)

        attn_output = dot_product_attention(
            "position",
            "key_position",
            "head_size",
            q,
            k,
            v,
            mask=mask,
            inference=self.inference,
            use_flash=self.config.use_flash_attention,
            attn_backend=self.config.attn_backend,
            flash_block_size=self.config.flash_attention_block_size,
            prng=k_drop,
            attention_dtype=jnp.float32 if self.config.upcast_attn else None,
        )

        attn_output = attn_output.astype(x.dtype)
        attn_output = self.c_proj(attn_output, key=k_out)

        return attn_output


class Gpt2Block(eqx.Module):
    ln_1: hnn.LayerNorm
    attn: Gpt2Attention
    ln_2: hnn.LayerNorm
    mlp: Gpt2Mlp
    resid_dropout: hnn.Dropout

    @staticmethod
    def init(config: Gpt2Config, *, key) -> "Gpt2Block":
        k_attn, k_mlp = jrandom.split(key, 2)

        ln_1 = hnn.LayerNorm.init(config.Embed, eps=config.layer_norm_epsilon, use_bias=config.use_bias)
        attn = Gpt2Attention.init(config, key=k_attn)
        ln_2 = hnn.LayerNorm.init(config.Embed, eps=config.layer_norm_epsilon, use_bias=config.use_bias)
        mlp = Gpt2Mlp.init(config.Embed, config.Mlp, config.activation_function, key=k_mlp, use_bias=config.use_bias)
        resid_dropout = hnn.Dropout(pdrop=config.resid_pdrop)

        return Gpt2Block(ln_1, attn, ln_2, mlp, resid_dropout)

    @named_call
    def __call__(self, x: NamedArray, mask: Optional[AttentionMask | NamedArray], layer_idx, *, key):
        k1, k2, k3, k4 = haliax.jax_utils.maybe_rng_split(key, 4)

        attn_output = self.attn(self.ln_1(x), mask=mask, layer_idx=layer_idx, key=k1)
        attn_output = self.resid_dropout(attn_output, key=k2)
        x = x + attn_output

        ff_output = self.mlp(self.ln_2(x), key=k3)
        ff_output = self.resid_dropout(ff_output, key=k4)
        x = x + ff_output

        return x


class Gpt2Transformer(ModuleWithStateDictSerialization):
    config: Gpt2Config = eqx.static_field()
    blocks: Stacked[Gpt2Block]
    ln_f: hnn.LayerNorm

    @staticmethod
    def init(config: Gpt2Config, *, key):
        # vectorize the blocks
        blocks = Stacked.init(config.Layers, Gpt2Block, gradient_checkpointing=config.gradient_checkpointing)(
            config,
            key=shaped_rng_split(key, config.num_layers),
        )
        ln_f = hnn.LayerNorm.init(config.Embed, eps=config.layer_norm_epsilon, use_bias=config.use_bias)

        return Gpt2Transformer(config, blocks, ln_f)

    @named_call
    def __call__(self, x: NamedArray, attn_mask: Optional[AttentionMask | NamedArray], *, key=None) -> NamedArray:
        keys = hax.jax_utils.maybe_rng_split(key, self.config.num_layers) if key is not None else None
        x = self.blocks.fold(x, attn_mask, hax.arange(self.config.Layers), key=keys)
        x = self.ln_f(x)

        return x

    def _state_dict_key_map(self) -> Dict[str, Optional[str]]:
        return {"blocks": "h"}


class Gpt2Embeddings(ModuleWithStateDictSerialization, eqx.Module):
    Vocab: Axis = eqx.static_field()
    config: Gpt2Config = eqx.static_field()

    token_embeddings: hnn.Embedding
    position_embeddings: hnn.Embedding
    dropout: hnn.Dropout

    @staticmethod
    def init(Vocab: Axis, config: Gpt2Config, *, key) -> "Gpt2Embeddings":
        k_wte, k_wpe, k_out = jrandom.split(key, 3)

        token_embeddings = hnn.Embedding.init(
            Vocab, config.Embed, key=k_wte, initializer_range=config.initializer_range
        )
        position_embeddings = hnn.Embedding.init(
            config.Pos, config.Embed, key=k_wpe, initializer_range=config.initializer_range / 2
        )
        dropout = hnn.Dropout(pdrop=config.embed_pdrop)

        return Gpt2Embeddings(Vocab, config, token_embeddings, position_embeddings, dropout)

    @named_call
    def embed(self, input_ids, *, key):
        input_embeds = self.token_embeddings(input_ids)
        input_Pos = input_ids.resolve_axis("position")
        position_embeds = self.position_embeddings.embed(hax.arange(input_Pos))
        x = input_embeds + position_embeds
        x = self.dropout(x, key=key)

        return x

    def unembed(self, x: NamedArray):
        return hax.dot(x, self.token_embeddings.weight, axis="embed")

    def _state_dict_key_map(self) -> Dict[str, Optional[str]]:
        return {"token_embeddings": "wte", "position_embeddings": "wpe"}

    def resize_embeddings(self, new_size: int, key: Optional[PRNGKeyArray] = None):
        new_token_embeddings = self.token_embeddings.resize_embeddings(new_size, key=key)
        return dataclasses.replace(self, Vocab=self.Vocab.resize(new_size), token_embeddings=new_token_embeddings)


class Gpt2LMHeadModel(LmWithHfSerializationMixin[Gpt2Config]):
    transformer: Gpt2Transformer
    embeddings: Gpt2Embeddings

    @property
    def config(self):
        return self.transformer.config

    @property
    def Vocab(self) -> Axis:
        return self.embeddings.Vocab

    @property
    def Pos(self) -> Axis:
        return self.config.Pos

    @classmethod
    def init(cls, Vocab: Axis, config: Gpt2Config, *, key) -> "Gpt2LMHeadModel":
        k_t, k_embeddings = jrandom.split(key, 2)
        transformer = Gpt2Transformer.init(config, key=k_t)
        embeddings = Gpt2Embeddings.init(Vocab, config, key=k_embeddings)

        return Gpt2LMHeadModel(transformer, embeddings)

    def activations(
        self, input_ids: NamedArray, attn_mask: Optional[AttentionMask | NamedArray] = None, *, key=None
    ) -> NamedArray:
        k_embed, k_transformer = haliax.jax_utils.maybe_rng_split(key, 2)
        x = self.embeddings.embed(input_ids, key=k_embed)
        x = self.transformer(x, attn_mask, key=k_transformer)

        return x

    def get_lm_head(self) -> hax.NamedArray:
        return self.embeddings.token_embeddings.weight

    def resize_vocab(self, new_size: int, key: Optional[PRNGKeyArray] = None) -> "Gpt2LMHeadModel":
        new_embeddings = self.embeddings.resize_embeddings(new_size, key=key)
        return dataclasses.replace(self, embeddings=new_embeddings)

    def _state_dict_key_map(self) -> Dict[str, Optional[str]]:
        return {"transformer": None, "embeddings": None}


ACT2FN: Dict[str, Callable] = {
    "relu": hnn.relu,
    "silu": hnn.silu,
    "swish": hnn.swish,
    "gelu": partial(hnn.gelu, approximate=False),
    "gelu_new": partial(hnn.gelu, approximate=True),
    "quick_gelu": hnn.quick_gelu,
}
